/*
 * search.c
 *
 *      Revision 0.2   1996/01/15      H.Takenaka
 *
 */

#include	"xdvi.h"
#include	<ctype.h>
#include	<stdlib.h>
#include	<X11/keysym.h>

#ifndef	SEEK_SET
#define	SEEK_SET	0
#define	SEEK_CUR	1
#define	SEEK_END	2
#endif

static int	pages_allocated = 0;
static int	*text_allocated = NULL;
static int	*text_used = NULL;

#ifdef	PAGENUM
#define	PAGE_OFFSET(x)	page_index[x].offset
#else
#define	PAGE_OFFSET(x)	page_offset[x]
#endif

/*
 * Default Font Encodings
 */

#define	FONTENC_NONE	-1
#define	FONTENC_ALNUM	0
#define	FONTENC_TEXTEXT	1
#define	FONTENC_TEXMATHSYM	2
#define	FONTENC_TEXWOLIG	3
#define	FONTENC_TEXTT	4
#define	FONTENC_TEXEXT	5
#define	FONTENC_CORK	6
#define	FONTENC_DNP		0x1000

#define	HASH_SIZE	32

static struct coding_schemes {
	struct coding_schemes *next;
	char	*name;
	int		index;
} CodingSchemes[] = {
	{ NULL, "AEFMNOPST ONLY", FONTENC_ALNUM },
	{ NULL, "ASCII CAPS AND DIGITS", FONTENC_ALNUM },
	{ NULL, "EULER SUBSTITUTIONS ONLY", FONTENC_NONE },
	{ NULL, "EXTENDED TEX FONT ENCODING - LATIN", FONTENC_CORK },
	{ NULL, "LATEX CIRCLE", FONTENC_NONE },
	{ NULL, "LATEX LINE", FONTENC_NONE },
	{ NULL, "LATEX SYMBOLS", FONTENC_NONE },
	{ NULL, "TEX EXTENDED ASCII", FONTENC_TEXTEXT },
	{ NULL, "TEX MATH EXTENSION", FONTENC_NONE },
	{ NULL, "TEX MATH ITALIC SUBSET", FONTENC_ALNUM },
	{ NULL, "TEX MATH ITALIC", FONTENC_ALNUM },
	{ NULL, "TEX MATH SYMBOLS SUBSET", FONTENC_TEXMATHSYM },
	{ NULL, "TEX MATH SYMBOLS", FONTENC_TEXMATHSYM },
	{ NULL, "TEX TEXT + ADOBESTANDARDENCODING", FONTENC_ALNUM },
	{ NULL, "TEX TEXT SUBSET", FONTENC_TEXTEXT },
	{ NULL, "TEX TEXT WITHOUT F-LIGATURES", FONTENC_TEXWOLIG },
	{ NULL, "TEX TEXT", FONTENC_TEXTEXT },
	{ NULL, "TEX TYPEWRITER TEXT", FONTENC_TEXTT },
	{ NULL, "UNSPECIFIED", FONTENC_NONE },
	{ NULL, "DNPKANJI SY", FONTENC_DNP + 1 },
	{ NULL, "DNPKANJI ROMA", FONTENC_DNP + 2 },
	{ NULL, "DNPKANJI HIRA", FONTENC_DNP + 3 },
	{ NULL, "DNPKANJI KATA", FONTENC_DNP + 4 },
	{ NULL, "DNPKANJI GREEK", FONTENC_DNP + 5 },
	{ NULL, "DNPKANJI RUSSIAN", FONTENC_DNP + 6 },
	{ NULL, "DNPKANJI KEISEN", FONTENC_DNP + 7 },
	{ NULL, "DNPKANJI KA", FONTENC_DNP + 8 },
	{ NULL, "DNPKANJI KB", FONTENC_DNP + 9 },
	{ NULL, "DNPKANJI KC", FONTENC_DNP + 10 },
	{ NULL, "DNPKANJI KD", FONTENC_DNP + 11 },
	{ NULL, "DNPKANJI KE", FONTENC_DNP + 12 },
	{ NULL, "DNPKANJI KF", FONTENC_DNP + 13 },
	{ NULL, "DNPKANJI KG", FONTENC_DNP + 14 },
	{ NULL, "DNPKANJI KH", FONTENC_DNP + 15 },
	{ NULL, "DNPKANJI KI", FONTENC_DNP + 16 },
	{ NULL, "DNPKANJI KJ", FONTENC_DNP + 17 },
	{ NULL, "DNPKANJI KK", FONTENC_DNP + 18 },
	{ NULL, "DNPKANJI KL", FONTENC_DNP + 19 },
	{ NULL, "DNPKANJI KM", FONTENC_DNP + 20 },
	{ NULL, "DNPKANJI KN", FONTENC_DNP + 21 },
	{ NULL, "DNPKANJI KO", FONTENC_DNP + 22 },
	{ NULL, "DNPKANJI KP", FONTENC_DNP + 23 },
	{ NULL, "DNPKANJI KQ", FONTENC_DNP + 24 },
	{ NULL, "DNPKANJI KR", FONTENC_DNP + 25 },
	{ NULL, "DNPKANJI KS", FONTENC_DNP + 26 },
	{ NULL, "DNPKANJI KT", FONTENC_DNP + 27 },
	{ NULL, "DNPKANJI KU", FONTENC_DNP + 28 },
	{ NULL, "DNPKANJI KV", FONTENC_DNP + 29 },
	{ NULL, "DNPKANJI KW", FONTENC_DNP + 30 },
	{ NULL, "DNPKANJI KX", FONTENC_DNP + 31 },
	{ NULL, "DNPKANJI KY", FONTENC_DNP + 32 },
	{ NULL, "DNPKANJI KZ", FONTENC_DNP + 33 },
	{ NULL, NULL, 0 }
};

static struct coding_schemes *HashCodingSchemes[HASH_SIZE];


static struct default_encoding {
	char	*name;
	char	*coding;
} Font2CodingScheme[] = {
	{ "dmjsy", "DNPKANJI SY" },	{ "dmjroma", "DNPKANJI ROMA" },
	{ "dmjhira","DNPKANJI HIRA" },	{ "dmjkata", "DNPKANJI KATA" },
	{ "dmjgreek", "DNPKANJI GREEK" },	{ "dmjrussian", "DNPKANJI RUSSIAN" },
	{ "dmjkeisen", "DNPKANJI KEISEN" },	{ "dmjka", "DNPKANJI KA" },
	{ "dmjkb", "DNPKANJI KB" },	{ "dmjkc", "DNPKANJI KC" },
	{ "dmjkd", "DNPKANJI KD" },	{ "dmjke", "DNPKANJI KE" },
	{ "dmjkf", "DNPKANJI KF" },	{ "dmjkg", "DNPKANJI KG" },
	{ "dmjkh", "DNPKANJI KH" },	{ "dmjki", "DNPKANJI KI" },
	{ "dmjkj", "DNPKANJI KJ" },	{ "dmjkk", "DNPKANJI KK" },
	{ "dmjkl", "DNPKANJI KL" },	{ "dmjkm", "DNPKANJI KM" },
	{ "dmjkn", "DNPKANJI KN" },	{ "dmjko", "DNPKANJI KO" },
	{ "dmjkp", "DNPKANJI KP" },	{ "dmjkq", "DNPKANJI KQ" },
	{ "dmjkr", "DNPKANJI KR" },	{ "dmjks", "DNPKANJI KS" },
	{ "dmjkt", "DNPKANJI KT" },	{ "dmjku", "DNPKANJI KU" },
	{ "dmjkv", "DNPKANJI KV" },	{ "dmjkw", "DNPKANJI KW" },
	{ "dmjkx", "DNPKANJI KX" },	{ "dmjky", "DNPKANJI KY" },
	{ "dmjkz", "DNPKANJI KZ" },

	{ "dgjsy", "DNPKANJI SY" },	{ "dgjroma", "DNPKANJI ROMA" },
	{ "dgjhira","DNPKANJI HIRA" },	{ "dgjkata", "DNPKANJI KATA" },
	{ "dgjgreek", "DNPKANJI GREEK" },	{ "dgjrussian", "DNPKANJI RUSSIAN" },
	{ "dgjkeisen", "DNPKANJI KEISEN" },	{ "dgjka", "DNPKANJI KA" },
	{ "dgjkb", "DNPKANJI KB" },	{ "dgjkc", "DNPKANJI KC" },
	{ "dgjkd", "DNPKANJI KD" },	{ "dgjke", "DNPKANJI KE" },
	{ "dgjkf", "DNPKANJI KF" },	{ "dgjkg", "DNPKANJI KG" },
	{ "dgjkh", "DNPKANJI KH" },	{ "dgjki", "DNPKANJI KI" },
	{ "dgjkj", "DNPKANJI KJ" },	{ "dgjkk", "DNPKANJI KK" },
	{ "dgjkl", "DNPKANJI KL" },	{ "dgjkm", "DNPKANJI KM" },
	{ "dgjkn", "DNPKANJI KN" },	{ "dgjko", "DNPKANJI KO" },
	{ "dgjkp", "DNPKANJI KP" },	{ "dgjkq", "DNPKANJI KQ" },
	{ "dgjkr", "DNPKANJI KR" },	{ "dgjks", "DNPKANJI KS" },
	{ "dgjkt", "DNPKANJI KT" },	{ "dgjku", "DNPKANJI KU" },
	{ "dgjkv", "DNPKANJI KV" },	{ "dgjkw", "DNPKANJI KW" },
	{ "dgjkx", "DNPKANJI KX" },	{ "dgjky", "DNPKANJI KY" },
	{ "dgjkz", "DNPKANJI KZ" },

	{ "cmb", "TEX TEXT" },	{ "cmbsy", "TEX MATH SYMBOLS" },
	{ "cmbx", "TEX TEXT" },	{ "cmbxsl", "TEX TEXT" },
	{ "cmbxti", "TEX TEXT" },	{ "cmcsc", "TEX TEXT WITHOUT F-LIGATURES" },
	{ "cmdunh", "TEX TEXT" },	{ "cmex", "TEX MATH EXTENSION" },
	{ "cmff", "TEX TEXT" },	{ "cmfi", "TEX TEXT" },
	{ "cmfib", "TEX TEXT" },	{ "cmitt", "TEX TYPEWRITER TEXT" },
	{ "cmmi", "TEX MATH ITALIC" },	{ "cmmib", "TEX MATH ITALIC" },
	{ "cmr", "TEX TEXT" },	{ "cmsl", "TEX TEXT" },
	{ "cmsltt", "TEX TYPEWRITER TEXT" },	{ "cmss", "TEX TEXT" },
	{ "cmssbx", "TEX TEXT" },	{ "cmssdc", "TEX TEXT" },
	{ "cmssi", "TEX TEXT" },	{ "cmssq", "TEX TEXT" },
	{ "cmsy", "TEX MATH SYMBOLS" },	{ "cmtcsc", "TEX TYPEWRITER TEXT" },
	{ "cmtex", "TEX EXTENDED ASCII" },	{ "cmti", "TEX TEXT" },
	{ "cmtt", "TEX TYPEWRITER TEXT" },	{ "cmu", "TEX TEXT" },
	{ "cmvtt", "TEX TEXT" },	{ "cminch" , "ASCII CAPS AND DIGITS" },

	{ "cccsc", "TEX TEXT WITHOUT F-LIGATURES" }, { "ccmi", "TEX MATH ITALIC" },
	{ "ccr", "TEX TEXT" }, { "ccsl", "TEX TEXT" },
	{ "ccslc", "TEX TEXT" }, { "ccti", "TEX TEXT" },
	{ "dcb", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dcbx", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dcbxsl", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dcbxti", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dccsc", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dcdunh", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dcff", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dcfi", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dcfib", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dcitt", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dcr", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dcsl", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dcsltt", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dcss", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dcssbx", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dcssi", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dctcsc", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dcti", "CODINGSCHEME EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dctt", "CODINGSCHEME EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dcu", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "dcvtt", "EXTENDED TEX FONT ENCODING - LATIN" },
	{ "euex", "EULER SUBSTITUTIONS ONLY" },
	{ "eufb", "TEX TEXT SUBSET" },
	{ "eufm", "TEX TEXT SUBSET" },
	{ "eurb", "TEX MATH ITALIC SUBSET" },
	{ "eurm", "TEX MATH ITALIC SUBSET" },
	{ "eusb", "TEX MATH SYMBOLS SUBSET" },
	{ "eusm", "TEX MATH SYMBOLS SUBSET" },
	{ "icmcsc", "TEX TEXT WITHOUT F-LIGATURES" },
	{ "icmex", "TEX MATH EXTENSION" },
	{ "icmmi", "TEX MATH ITALIC" },
	{ "icmsy", "TEX MATH SYMBOLS" },
	{ "icmtt", "TEX TYPEWRITER TEXT" },
	{ "ilasy", "LATEX SYMBOLS" },
	{ "ilcmss", "TEX TEXT" },
	{ "ilcmssb", "TEX TEXT" },
	{ "ilcmssi", "TEX TEXT" },
	{ "lasy", "LATEX SYMBOLS" },
	{ "lasyb", "LATEX SYMBOLS" },
	{ "lcircle", "LATEX CIRCLE" },
	{ "lcirclew", "LATEX CIRCLE" },
	{ "lcmss", "TEX TEXT" },
	{ "lcmssb", "TEX TEXT" },
	{ "lcmssi", "TEX TEXT" },
	{ "line", "LATEX LINE" },
	{ "linew", "LATEX LINE" },
	{ "logo", "AEFMNOPST ONLY" },
	{ "logobf", "AEFMNOPST ONLY" },
	{ "logod", "AEFMNOPST ONLY" },
	{ "logosl", "AEFMNOPST ONLY" },
	{ "msam", "TEX MATH SYMBOLS" },
	{ "msbm", "TEX MATH SYMBOLS" },

	{ NULL, NULL }
};

static struct font2codingscheme {
	struct font2codingscheme *next;
	unsigned char	*name;
	int		coding;
} *HashFont2CodingScheme[HASH_SIZE];


static struct coding_convert {
	char	*table[256];
} **CodingConvert = NULL;

static int coding_convert_size = 0;

#define	DEFAULT_CODINGCONV	(FONTENC_CORK+1)
static struct coding_convert DefaultCodingConvert[DEFAULT_CODINGCONV] = {
/* FONTENC_ALNUM */
	{	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 000 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 010 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 020 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 030 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 040 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 050 */
		 "0",  "1",  "2",  "3",  "4",  "5",  "6",  "7", /* 060 */
		 "8",  "9", NULL, NULL, NULL, NULL, NULL, NULL, /* 070 */
		NULL,  "A",  "B",  "C",  "D",  "E",  "F",  "G", /* 100 */
		 "H",  "I",  "J",  "K",  "L",  "M",  "N",  "O", /* 110 */
		 "P",  "Q",  "R",  "S",  "T",  "U",  "V",  "W", /* 120 */
		 "X",  "Y",  "Z", NULL, NULL, NULL, NULL, NULL, /* 130 */
		NULL,  "a",  "b",  "c",  "d",  "e",  "f",  "g", /* 140 */
		 "h",  "i",  "j",  "k",  "l",  "m",  "n",  "o", /* 150 */
		 "p",  "q",  "r",  "s",  "t",  "u",  "v",  "w", /* 160 */
		 "x",  "y",  "z", NULL, NULL, NULL, NULL, NULL, /* 170 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 200 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 210 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 220 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 230 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 240 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 250 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 260 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 270 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 300 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 310 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 320 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 330 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 340 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 350 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 360 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL },/* 370 */
/* FONTENC_TEXTEXT */
	{	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 000 */
		NULL, NULL, NULL, "ff", "fi", "fl", "ffi", "ffl", /* 010 */
		 "i",  "j",  "`",  "'",  "v",  "u",  "=",  "a", /* 020 */
		 "c", "ss", "ae", "oe",  "o", "AE", "OE",  "O", /* 030 */
		NULL,  "!", "''",  "#",  "$",  "%",  "&",  "'", /* 040 */
		 "(",  ")",  "*",  "+",  ",", NULL,  ".",  "/", /* 050 */
		 "0",  "1",  "2",  "3",  "4",  "5",  "6",  "7", /* 060 */
		 "8",  "9",  ":",  ";",  "!",  "=",  "?",  "?", /* 070 */
		 "@",  "A",  "B",  "C",  "D",  "E",  "F",  "G", /* 100 */
		 "H",  "I",  "J",  "K",  "L",  "M",  "N",  "O", /* 110 */
		 "P",  "Q",  "R",  "S",  "T",  "U",  "V",  "W", /* 120 */
		 "X",  "Y",  "Z",  "[",  "``",  "]", "^", NULL, /* 130 */
		 "`",  "a",  "b",  "c",  "d",  "e",  "f",  "g", /* 140 */
		 "h",  "i",  "j",  "k",  "l",  "m",  "n",  "o", /* 150 */
		 "p",  "q",  "r",  "s",  "t",  "u",  "v",  "w", /* 160 */
		 "x",  "y",  "z", "--", "---", "\"", "~", "\"", /* 170 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 200 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 210 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 220 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 230 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 240 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 250 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 260 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 270 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 300 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 310 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 320 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 330 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 340 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 350 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 360 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL },/* 370 */
/* FONTENC_TEXMATHSYM */
	{	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 000 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 010 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 020 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 030 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 040 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 050 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 060 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 070 */
		NULL,  "A",  "B",  "C",  "D",  "E",  "F",  "G", /* 100 */
		 "H",  "I",  "J",  "K",  "L",  "M",  "N",  "O", /* 110 */
		 "P",  "Q",  "R",  "S",  "T",  "U",  "V",  "W", /* 120 */
		 "X",  "Y",  "Z", NULL, NULL, NULL, NULL, NULL, /* 130 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 140 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 150 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 160 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 170 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 200 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 210 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 220 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 230 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 240 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 250 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 260 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 270 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 300 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 310 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 320 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 330 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 340 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 350 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 360 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL },/* 370 */
/* FONTENC_TEXWOLIG */
	{	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 000 */
		NULL, NULL, NULL, NULL, NULL, NULL,  "!",  "?", /* 010 */
		 "i",  "j",  "`",  "'",  "v",  "u",  "=",  "a", /* 020 */
		 "c", "ss", "ae", "oe",  "o", "AE", "OE",  "O", /* 030 */
		NULL,  "!", "''",  "#",  "$",  "%",  "&",  "'", /* 040 */
		 "(",  ")",  "*",  "+",  ",", NULL,  ".",  "/", /* 050 */
		 "0",  "1",  "2",  "3",  "4",  "5",  "6",  "7", /* 060 */
		 "8",  "9",  ":",  ";",  "<",  "=",  ">",  "?", /* 070 */
		 "@",  "A",  "B",  "C",  "D",  "E",  "F",  "G", /* 100 */
		 "H",  "I",  "J",  "K",  "L",  "M",  "N",  "O", /* 110 */
		 "P",  "Q",  "R",  "S",  "T",  "U",  "V",  "W", /* 120 */
		 "X",  "Y",  "Z",  "[",  "``",  "]", "^", NULL, /* 130 */
		 "`",  "a",  "b",  "c",  "d",  "e",  "f",  "g", /* 140 */
		 "h",  "i",  "j",  "k",  "l",  "m",  "n",  "o", /* 150 */
		 "p",  "q",  "r",  "s",  "t",  "u",  "v",  "w", /* 160 */
		 "x",  "y",  "z", "--", "---", "\"", "~", "\"", /* 170 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 200 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 210 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 220 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 230 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 240 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 250 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 260 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 270 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 300 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 310 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 320 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 330 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 340 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 350 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 360 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL },/* 370 */
/* FONTENC_TEXTT */
	{	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 000 */
		NULL, NULL, NULL, NULL, NULL, NULL,  "!",  "?", /* 010 */
		 "i",  "j",  "`",  "'",  "v",  "u",  "=",  "a", /* 020 */
		 "c", "ss", "ae", "oe",  "o", "AE", "OE",  "O", /* 030 */
		NULL,  "!", "\"",  "#",  "$",  "%",  "&",  "'", /* 040 */
		 "(",  ")",  "*",  "+",  ",", NULL,  ".",  "/", /* 050 */
		 "0",  "1",  "2",  "3",  "4",  "5",  "6",  "7", /* 060 */
		 "8",  "9",  ":",  ";",  "<",  "=",  ">",  "?", /* 070 */
		 "@",  "A",  "B",  "C",  "D",  "E",  "F",  "G", /* 100 */
		 "H",  "I",  "J",  "K",  "L",  "M",  "N",  "O", /* 110 */
		 "P",  "Q",  "R",  "S",  "T",  "U",  "V",  "W", /* 120 */
		 "X",  "Y",  "Z",  "[",  "\\",  "]", "^",  "_", /* 130 */
		 "`",  "a",  "b",  "c",  "d",  "e",  "f",  "g", /* 140 */
		 "h",  "i",  "j",  "k",  "l",  "m",  "n",  "o", /* 150 */
		 "p",  "q",  "r",  "s",  "t",  "u",  "v",  "w", /* 160 */
		 "x",  "y",  "z",  "{",  "|",  "}",  "~", "\"", /* 170 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 200 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 210 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 220 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 230 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 240 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 250 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 260 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 270 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 300 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 310 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 320 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 330 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 340 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 350 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 360 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL },/* 370 */
/* FONTENC_TEXEXT */
	{	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 000 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 010 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 020 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 030 */
		NULL,  "!", "\"",  "#",  "$",  "%",  "&",  "'", /* 040 */
		 "(",  ")",  "*",  "+",  ",", NULL,  ".",  "/", /* 050 */
		 "0",  "1",  "2",  "3",  "4",  "5",  "6",  "7", /* 060 */
		 "8",  "9",  ":",  ";",  "<",  "=",  ">",  "?", /* 070 */
		 "@",  "A",  "B",  "C",  "D",  "E",  "F",  "G", /* 100 */
		 "H",  "I",  "J",  "K",  "L",  "M",  "N",  "O", /* 110 */
		 "P",  "Q",  "R",  "S",  "T",  "U",  "V",  "W", /* 120 */
		 "X",  "Y",  "Z",  "[",  "\\",  "]", "^",  "_", /* 130 */
		 "`",  "a",  "b",  "c",  "d",  "e",  "f",  "g", /* 140 */
		 "h",  "i",  "j",  "k",  "l",  "m",  "n",  "o", /* 150 */
		 "p",  "q",  "r",  "s",  "t",  "u",  "v",  "w", /* 160 */
		 "x",  "y",  "z",  "{",  "|",  "}",  "~", NULL, /* 170 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 200 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 210 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 220 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 230 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 240 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 250 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 260 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 270 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 300 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 310 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 320 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 330 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 340 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 350 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, /* 360 */
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL },/* 370 */
/* FONTENC_CORK */
	{	 "`",  "'",  "^",  "~", NULL, NULL,  "a",  "v", /* 000 */
		 "u",  "=", NULL,  "c", NULL,  ",",  "<",  ">", /* 010 */
		"``", "''", ",,", "<<", ">>", "--", "---", NULL, /* 020 */
		NULL,  "i",  "j", "ff", "fi", "fl", "ffi", "ffl", /* 030 */
		NULL,  "!", "\"",  "#",  "$",  "%",  "&",  "'", /* 040 */
		 "(",  ")",  "*",  "+",  ",", NULL,  ".",  "/", /* 050 */
		 "0",  "1",  "2",  "3",  "4",  "5",  "6",  "7", /* 060 */
		 "8",  "9",  ":",  ";",  "<",  "=",  ">",  "?", /* 070 */
		 "@",  "A",  "B",  "C",  "D",  "E",  "F",  "G", /* 100 */
		 "H",  "I",  "J",  "K",  "L",  "M",  "N",  "O", /* 110 */
		 "P",  "Q",  "R",  "S",  "T",  "U",  "V",  "W", /* 120 */
		 "X",  "Y",  "Z",  "[",  "\\",  "]", "^",  "_", /* 130 */
		 "`",  "a",  "b",  "c",  "d",  "e",  "f",  "g", /* 140 */
		 "h",  "i",  "j",  "k",  "l",  "m",  "n",  "o", /* 150 */
		 "p",  "q",  "r",  "s",  "t",  "u",  "v",  "w", /* 160 */
		 "x",  "y",  "z",  "{",  "|",  "}",  "~", NULL, /* 170 */
		"\200", "\201", "\202", "\203", "\204", "\205", "\206", "\207",
		"\210", "\211", "\212", "\213", "\214", "\215", "\216", "\217",
		"\220", "\221", "\222", "\223", "\224", "\225", "\226", "\227",
		"\230", "\231", "\232", "\233", "\234", "\235", "\236", "\237",
		"\240", "\241", "\242", "\243", "\244", "\245", "\246", "\247",
		"\250", "\251", "\252", "\253", "\254", "\255", "\256", "\257",
		"\260", "\261", "\262", "\263", "\264", "\265", "\266", "\267",
		"\270", "\271", "\272", "\273", "\274", "\275", "\276", "\277",
		"\300", "\301", "\302", "\303", "\304", "\305", "\306", "\307",
		"\310", "\311", "\312", "\313", "\314", "\315", "\316", "\317",
		"\320", "\321", "\322", "\323", "\324", "\325", "\326", "\327",
		"\330", "\331", "\332", "\333", "\334", "\335", "\336", "\337",
		"\340", "\341", "\342", "\343", "\344", "\345", "\346", "\347",
		"\350", "\351", "\352", "\353", "\354", "\355", "\356", "\357",
		"\360", "\361", "\362", "\363", "\364", "\365", "\366", "\367",
		"\370", "\371", "\372", "\373", "\374", "\375", "\376", "\377"},
};

#undef HAVE_STRICMP
#undef HAVE_STRCASECMP

#ifdef HAVE_STRCASECMP
#define	stricmp(s1, s2)	strcasecmp(s1, s2)
#else
#ifndef HAVESTRICMP
#define	stricmp(s1, s2)	_stricmp(s1, s2)

int
_stricmp(s1, s2)
	_Xconst char	*s1;
	_Xconst char	*s2;
{
	while (*s1 && *s2) {
	    int i = tolower(*s1) - tolower(*s2);
	    if (i != 0) return i;
	    ++s1;
	    ++s2;
	}
	return tolower(*s1) - tolower(*s2);
}
#endif
#endif

static void *
xrealloc(ptr, size)
	void *ptr;
	unsigned size;
{
	if (ptr == NULL) return malloc(size);
	else return realloc(ptr, size);
}

int
get_hashed_codingscheme(csname)
	char *csname;
{
	char *p = csname;
	int	hash = 0;
	struct coding_schemes *ptr;

	while (*p) hash += toupper(*p++) - ' ';
	hash = ((hash * hash) / 128) % HASH_SIZE;

	for (ptr = HashCodingSchemes[hash]; ptr != NULL; ptr = ptr->next)
		if (stricmp(ptr->name, csname) == 0) {
			return ptr->index;
		}

	ptr = (struct coding_schemes *)xmalloc(sizeof(struct coding_schemes),
										   "search");
	ptr->name = strDup(csname);
	ptr->index = FONTENC_ALNUM;
	ptr->next = HashCodingSchemes[hash];
	HashCodingSchemes[hash] = ptr;

	return FONTENC_ALNUM;
}

/*
 * initialize tables.
 */
void
initialize_search(fc)
	_Xconst char *fc;
{
	char *p;
	int	i, hash;
	struct coding_schemes *ptr;
	struct default_encoding *ptr2;
	struct font2codingscheme *ptr3;

	CodingConvert =
		(struct coding_convert **) xmalloc(sizeof(struct coding_convert *)
										   * DEFAULT_CODINGCONV, "search");

	coding_convert_size = DEFAULT_CODINGCONV;

	for (i = 0; i < DEFAULT_CODINGCONV; i++)
		CodingConvert[i] = &DefaultCodingConvert[i];

	for (ptr = CodingSchemes; ptr->name != NULL; ptr++) {
		p = ptr->name;
		hash = 0;
		while (*p) hash += toupper(*p++) - ' ';
		hash = ((hash * hash) / 128) % HASH_SIZE;
		ptr->next = HashCodingSchemes[hash];
		HashCodingSchemes[hash] = ptr;
	}

/*
 *  Codes to load user-defined font-encodings from file <fc>
 *  must be here, but not implemented yet.
 */

	for (ptr2 = Font2CodingScheme; ptr2->name != NULL; ptr2++) {
		p = ptr2->name;
		hash = 0;
		while (*p) hash += *p++ - ' ';
		hash = ((hash * hash) / 32) % HASH_SIZE;
		ptr3 = (struct font2codingscheme *)
			xmalloc(sizeof(struct font2codingscheme), "search");
		ptr3->next = HashFont2CodingScheme[hash];
		ptr3->name = strDup(ptr2->name);
		HashFont2CodingScheme[hash] = ptr3;
		ptr3->coding = get_hashed_codingscheme(ptr2->coding);
	}
}

int
get_coding_scheme(fname)
	char *fname;
{
	FILE	*fp;
	char	*p, *p2;
	unsigned char	buf[BUFSIZ];
	int		i, hash = 0;
	struct font2codingscheme *ptr;

	p = buf;
	p2 = fname;
	while (*p2 && !isdigit(*p2)) {
		hash += *p2 - ' ';
		*p++ = *p2++;
	}
	*p = '\0';
	hash = ((hash * hash) / 32) % HASH_SIZE;

	for (ptr = HashFont2CodingScheme[hash]; ptr != NULL; ptr = ptr->next) {
		if (strcmp(ptr->name, buf) == 0) return ptr->coding;
	}
	ptr = (struct font2codingscheme *)
   		     xmalloc(sizeof(struct font2codingscheme), "search");
	ptr->name = strDup(buf);
	ptr->next = HashFont2CodingScheme[hash];
	HashFont2CodingScheme[hash] = ptr;
	fp = open_tfm(fname, &p2);
	if (fp == NULL) {
		Fprintf(stderr, "Can't open %s.tfm\n", fname);
		ptr->coding = FONTENC_ALNUM;
		return FONTENC_ALNUM;
	}
	free(p2);

	/*
	 *	get header length
	 */
	Fseek(fp, (long)0, 0);
	if (fread(buf, sizeof(char), 24, fp) < 24) goto error_ret;

	i = buf[3] * 256 + buf[4];
	if (i < 11) goto error_ret; /* coding scheme info doesn't exit. */

	/*
	 * get coding scheme
	 */
	Fseek(fp, (long)32, 0);
	if (fread(buf, sizeof(char), 40, fp) < 40) goto error_ret;
	fclose(fp);

	buf[buf[0]+1] = '\0';

	if (buf[0]) ptr->coding = get_hashed_codingscheme(&buf[1]);
	else ptr->coding = FONTENC_ALNUM;

	return ptr->coding;

error_ret:
	fclose(fp);
	return FONTENC_ALNUM;
}

/*
 * allocate text buffer for each pages.
 */
void
prepare_search(void)
{
	int	i;

	if (search_text != NULL) {
		for (i = 0; i < pages_allocated; i++) free(search_text[i]);
		free(search_text);
		free(text_allocated);
		free(text_used);
	}

	search_text = (SChar_t **)xmalloc(sizeof(SChar_t *)
										 * (unsigned)total_pages, "search");
	for (i = 0; i < total_pages; i++) search_text[i] = NULL;
	pages_allocated = total_pages;

	text_allocated = (int *)xmalloc(sizeof(int) * (unsigned)total_pages,
								  "search");
	text_used = (int *)xmalloc(sizeof(int) * (unsigned)total_pages,
								  "search");
	for (i = 0; i < total_pages; i++) text_allocated[i] = 0;
	for (i = 0; i < total_pages; i++) text_used[i] = 0;

	lastcmd_was_search = 0;
}

void
append_text(ch, fontp)
	int ch;
	struct font *fontp;
{
	int	cs;
	unsigned char *ptr;

	if (text_used[current_page] + 1 >= text_allocated[current_page]) {
		search_text[current_page] =
			(SChar_t *)xrealloc(search_text[current_page],
				  sizeof(SChar_t) * (text_allocated[current_page] + 1024));
		text_allocated[current_page] += 1024;
	}

	cs = fontp->coding_scheme;
#if	defined(KANJI) && !defined(ASCIIDNP)
	if (fontp->flags & FONT_KANJI) {
		search_text[current_page][text_used[current_page]++] = ch;
	}
	else
#endif
	if (cs == FONTENC_NONE) return;
	else if (cs >= FONTENC_DNP)
		search_text[current_page][text_used[current_page]++] = 
			(SChar_t)ku_ten_compute(cs - FONTENC_DNP, ch);
	else if (cs < coding_convert_size
			 && (ptr = CodingConvert[cs]->table[ch]) != NULL) {
		while (*ptr) {
			if (text_used[current_page] + 1 >= text_allocated[current_page]) {
				search_text[current_page] =
					(SChar_t *)xrealloc(search_text[current_page],
								   sizeof(SChar_t)
								   * (text_allocated[current_page] + 1024));
				text_allocated[current_page] += 1024;
			}
			search_text[current_page][text_used[current_page]] = *ptr++;
			text_used[current_page]++;
		}
	}
	else return;
	search_text[current_page][text_used[current_page]] = 0;
}

void
debug_out(void)
{
	int	c, i;

	for (i = 0; c = search_text[current_page][i]; i++) {
		if (c & 0xff00)
			Fprintf(stderr, "\x1b$B%c%c\x1b(B", (c >> 8) & 0xff, c & 0xff);
		else if (c < 0x20 || c >= 0x7f)
			Fprintf(stderr, "[%x]", c);
		else
			Fprintf(stderr, "%c", c);
	}
	Fprintf(stderr, "\n");
}

/*
 * convert search string to internal form.
 */

static SChar_t *
to_internal(str, len)
	unsigned char *str;
	int *len;
{
	static SChar_t *buf = NULL;
	static int allocated = 0;
	int	off, japanese;

	if (str == NULL || *str == '\0') {
		off = 0;
		if (buf != NULL) while (buf[off] != 0) off++;
		*len = off;
		return buf;
	}

	off = 0;
	japanese = 0;
	while (*str) {
		if (off + 1 >= allocated) {
			allocated += 16;
			buf = (SChar_t *)xrealloc(buf, sizeof(SChar_t) * allocated);
		}
		if (*str == '\x1b') {
			if (str[1] == '$' && (str[2] == 'B' || str[2] == '@')) {
				japanese = 1;
				str += 3;
			}
			else if (str[1] == '(' && (str[2] == 'B' || str[2] == 'J')) {
				japanese = 0;
				str += 3;
			}
			else str++;
		}
		else if (japanese && *str > 0x20 && *str < 0x7f) {
			if (str[1] && str[1] > 0x20 && str[1] < 0x7f) {
				buf[off] = (*str << 8) + str[1];
				off++;
				str += 2;
			}
		}
		else if (tolower(*cmdline_jcode) == 'e'
				 && *str > 0xA0 && *str < 0xff) {
			if (str[1] && str[1] > 0xA0 && str[1] < 0xff) {
				buf[off] = ((*str & 0x7f) << 8) + (str[1] & 0x7f);
				off++;
				str += 2;
 			}
		}
		else if (tolower(*cmdline_jcode) == 's'
				 && (*str >= 0x81 && *str <= 0x9f
					 || *str >= 0xe0 && *str <= 0xfc)) {
			if (str[1]) {
				if (*str <= 0x9f) {
					if (str[1] < 0x9f) buf[off] = (*str * 2 - 0xf1) << 8;
					else buf[off] = (*str * 2 - 0xf0) << 8;
				} else {
					if (str[1] < 0x9f) buf[off] = (*str * 2 - 0x161) << 8;
					else buf[off] = (*str * 2 - 0x160) << 8;
				}
				if (str[1] < 0x7f) buf[off] += str[1] - 0x1f;
				else if (str[1] < 0x9f) buf[off] += str[1] - 0x20;
				else buf[off] += str[1] - 0x7e;
			}
		}
		else if (isspace(*str)) str++;
		else if (!japanese && *str == '-') {
			if (str[1] == '-')
				while (*str && *str == '-') {
					buf[off] = '-';
					off++;
					str++;
					if (off + 1 >= allocated) {
						allocated += 16;
						buf = (SChar_t *)
							xrealloc(buf, sizeof(SChar_t) * allocated);
					}
				}
			else str++;
		}
		else if (*str == '\\' && str[1]) {
			str++;
			buf[off] = *str;
			off++;
			str++;
		}
		else {
			buf[off] = *str;
			off++;
			str++;
		}
	}

	buf[off] = 0;
	*len = off;
	return buf;
}

static void
scan_page(page)
	int page;
{
	int	save_curpage = current_page;

	current_page = page;
	search_mode = 2;
	draw_page();
	current_page = save_curpage;
}

static int
match_string(ptr, len, page, offset)
	SChar_t *ptr;
	int len, page, offset;
{
	int	len2;

	if (page < 0 || page >= total_pages) return 0;

	if (search_text[page] == NULL) scan_page(page);

	do {
		if (offset >= text_used[page]) {
			offset -= text_used[page];
			page++;
			if (page >= total_pages) return 0;
			if (search_text[page] == NULL) scan_page(page);
		}
		len2 = text_used[page] - offset;
		if (len2 > len) len2 = len;
		if (memcmp(ptr, search_text[page] + offset, sizeof(SChar_t) * len2))
			return 0;
		ptr += len2;
		offset += len2;
		len -= len2;
	} while (len);

	return 1;
}

static int
match_string_case(ptr, len, page, offset)
	SChar_t *ptr;
	int len, page, offset;
{
	int	i, len2;

	if (page < 0 || page >= total_pages) return 0;

	if (search_text[page] == NULL) scan_page(page);

	do {
		if (offset >= text_used[page]) {
			offset -= text_used[page];
			page++;
			if (page >= total_pages) return 0;
			if (search_text[page] == NULL) scan_page(page);
		}
		len2 = text_used[page] - offset;
		if (len2 > len) len2 = len;
		for (i = 0; i < len2; i++) {
			if (ptr[i] < 0x80 && search_text[page][offset+i] < 0x80) {
				if (tolower(ptr[i]) != tolower(search_text[page][offset+i]))
					return 0;
			}
			else if (ptr[i] != search_text[page][offset+i]) return 0;
		}
		ptr += len2;
		offset += len2;
		len -= len2;
	} while (len);

	return 1;
}


int
search_forward(str, cs, start)
	char *str;
	int	cs, start;
{
	int		page;
	SChar_t	*ptr;
	int		len, offset;

	ptr = to_internal(str, &len);
	if (ptr == NULL || len == 0) return -1;

	page = start;
	for (; page < total_pages; page++) {
		if (search_text[page] == NULL) scan_page(page);
		for (offset = 0; offset < text_used[page]; offset++)
			if (cs) {
				if (match_string_case(ptr, len, page, offset)) goto found;
			}
			else if (match_string(ptr, len, page, offset)) goto found;
	}
	(void) lseek(fileno(dvi_file), PAGE_OFFSET(current_page), SEEK_SET);
	return -1;

found:
	(void) lseek(fileno(dvi_file), PAGE_OFFSET(current_page), SEEK_SET);
	return page;
}

int
search_backward(str, cs)
	char *str;
	int	cs;
{
	int		page;
	SChar_t	*ptr;
	int		len, offset;

	ptr = to_internal(str, &len);
	if (ptr == NULL || len == 0) return -1;

	for (page = current_page; page--;) {
		if (search_text[page] == NULL) scan_page(page);
		for (offset = text_used[page]; offset--; )
			if (cs) {
				if (match_string_case(ptr, len, page, offset)) goto found;
			}
			else if (match_string(ptr, len, page, offset)) goto found;
	}
	(void) lseek(fileno(dvi_file), PAGE_OFFSET(current_page), SEEK_SET);
	return -1;
found:
	(void) lseek(fileno(dvi_file), PAGE_OFFSET(current_page), SEEK_SET);
	return page;
}

/*
 * match string in a very sloppy way.
 */
static int
sloppy_match_string(ptr, len, page, offset)
	SChar_t *ptr;
	int len, page, offset;
{
	static int *buf = NULL;
	static int allocated = 0;
	int		i, off, len2, save_len, match_len;

	if (page < 0 || page >= total_pages) return 0;

	if (len < 8) return match_string(ptr, len, page, offset);

	if (allocated < len * 2) {
		allocated = len * 2;
		buf = (int *)xrealloc(buf, sizeof(int) * allocated);
	}

	save_len = len;
	match_len = len2 = len * 2;

	for (i = 0; i < match_len; i++) buf[i] = 0;
	off = 0;

	if (search_text[page] == NULL) scan_page(page);
	do {
		if (offset >= text_used[page]) {
			offset -= text_used[page];
			page++;
			if (page >= total_pages) return 0;
			if (search_text[page] == NULL) scan_page(page);
		}

		if (len2 < len + 2) {
			ptr++;
			len--;
		}
		if (*ptr == search_text[page][offset]) {
			for (i = off; i < off + save_len && i < match_len; i++)
				buf[i]++;
			ptr++;
			len--;
		}
		offset++;
		len2--;
	} while (len2 && len);

	for (i = save_len / 2; i < match_len - 1; i++)
		if (buf[i] >= save_len / 2 && buf[i] > buf[i+1])
			return i;

	return 0;
}

int
sloppy_search_forward(str)
	char *str;
{
	int		page;
	SChar_t	*ptr;
	int		len, offset, rc;

	ptr = to_internal(str, &len);
	if (ptr == NULL || len == 0) return -1;

	for (page = 0; page < total_pages; page++) {
		if (search_text[page] == NULL) scan_page(page);
		for (offset = 0; offset < text_used[page]; offset++)
			if (rc = sloppy_match_string(ptr, len, page, offset)) goto found;
	}
	(void) lseek(fileno(dvi_file), PAGE_OFFSET(current_page), SEEK_SET);
	return -1;

found:
	(void) lseek(fileno(dvi_file), PAGE_OFFSET(current_page), SEEK_SET);
	if (offset + rc >= text_used[page]) page++;
	return page;
}


/*
 * This function is a preliminary version
 * and needs to be completely rewritten (maybe with X Toolkit).
 */
static char *
get_search_string(char *prompt)
{
	static char buf[BUFSIZ];
	static char	trbuf[16];
	XEvent	event;
	int		pos, nbytes;
	int		wid, new_wid, prompt_len;
	KeySym	key;
	static XFontStruct	*fnt = NULL;

	strcpy(buf, prompt);
	prompt_len = pos = strlen(prompt);

	if (fnt == NULL)
	    if ((fnt = XLoadQueryFont(DISP, display_font)) == NULL) {
			Fprintf(stderr, "%s : display font \"%s\" not loadable\n",
					prog, display_font);
			exit(1);
	    }

	XSetFont(DISP, foreGC, fnt->fid);
	wid = XTextWidth(fnt, buf, prompt_len);
	XDrawImageString(DISP, mane.win, foreGC, 5, 5 + fnt->ascent,
					 buf, prompt_len);

	while (1) {
		XNextEvent(DISP, &event);
		if (event.type == KeyPress) {
			nbytes = XLookupString(&event.xkey, trbuf, 16, &key, NULL);
			if (nbytes == 1) {
				switch (key) {
                case XK_Escape:
					XClearArea(DISP, mane.win, 5, 5,
							   5 + wid, 5 + fnt->ascent + fnt->descent,
							   False);
					return NULL;
				case XK_Return:
					XClearArea(DISP, mane.win, 5, 5,
							   5 + wid, 5 + fnt->ascent + fnt->descent,
							   False);
					return buf + prompt_len;
				case XK_BackSpace:
				case XK_Delete:
					if (pos == prompt_len) XBell(DISP, 20);
					else pos--;
					break;
				default:
					if (key < 0x100 && pos < BUFSIZ - 1)
						buf[pos++] = trbuf[0];
					else
						continue;
				}
				buf[pos] = 0;
				new_wid = XTextWidth(fnt, buf, pos);
				if (new_wid < wid)
					XClearArea(DISP, mane.win, 5 + new_wid, 5,
							   5 + wid, 5 + fnt->ascent + fnt->descent, False);
				wid = new_wid;
				XDrawImageString(DISP, mane.win, foreGC,
								 5, 5 + fnt->ascent, buf, pos);
			}
		}
	}
}

int
search_forward_interactive()
{
	char	*str;

	if ((str = get_search_string("Forward Search:")) == NULL) return -1;
	lastcmd_was_search = 1;
	if (lastcmd_was_search && *str == '\0')
		return search_forward(str, 1, current_page + 1);
	else
		return search_forward(str, 1, current_page);
}

int
search_backward_interactive()
{
	char	*str;

	if ((str = get_search_string("Backward Search:")) == NULL) return -1;
	lastcmd_was_search = 1;
	return search_backward(str, 1);
}
